﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Drawing;
using System.Xml.Serialization;
using System.Xml;

namespace GPXLib
{
	public class TrailSummary
	{
		private string _gpxPath;
		private GPXLib.wptType _highPoint = new GPXLib.wptType();
		private GPXLib.wptType _lowPoint = new GPXLib.wptType();
		private decimal _totalSpan; // the total span in meters from lowest to highest point
		private decimal _sectionSpan; // total span div. by 5 for each colored section
		private double _totalDistance; // total distance from start trackpoint to end trackpoint
		private double _totalElevChange; // total elevation change based on diff between each trackpoint
		private double _totalAscent;
		private GPXLib.gpxType _gpxDataSet;
		private List<GPXLib.wptType> _trackPointArray = new List<GPXLib.wptType>();
		private List<GPXLib.wptType> _waypointArray = new List<GPXLib.wptType>();
		//private ArrayList _comparePoints = new ArrayList();
		private string _elevLegend;
		//protected string gpxPath;

		/// <summary>
		/// 
		///  The following waypoint reference taken from Google maps, http://code.google.com/p/waypoints/wiki/Design
		/// 'lat' - Latitude (attribute) - The latitude of the point. Decimal degrees, WGS84 datum.
		/// 'lon' - Longitude (attribute) - The latitude of the point. Decimal degrees, WGS84 datum. Position Informatioin /// 
		/// 'ele' - Elevation (in meters) of the point. /// 
		/// 'time' - Creation/modification timestamp for element. Date and time in are in Univeral Coordinated Time (UTC), not local time! Conforms to ISO 8601 specification for date/time representation. Fractional seconds are allowed for millisecond timing in tracklogs. ///
		/// 'magvar' - Magnetic variation (in degrees) at the point /// 
		/// 'geoidheight' - Height (in meters) of geoid (mean sea level) above WGS84 earth ellipsoid. As defined in NMEA GGA message. Description Information /// 'name' - The GPS name of the waypoint. This field will be transferred to and from the GPS. GPX does not place restrictions on the length of this field or the characters contained in it. It is up to the receiving application to validate the field before sending it to the GPS. 
		/// 'cmt' - GPS waypoint comment. Sent to GPS as comment.
		/// 'desc' - A text description of the element. Holds additional information about the element intended for the user, not the GPS.
		/// 'src' - Source of data. Included to give user some idea of reliability and accuracy of data. "Garmin eTrex", "USGS quad Boston North", e.g. link - Link to additional information about the waypoint.
		/// 'url' - URL (GPX 1.0 only) /// 'urlname' - Name of URL (GPX 1.0 only)
		/// 'link' - Link to additional information about the waypoint. (GPX 1.1 only)
		/// 'sym' - Text of GPS symbol name. For interchange with other programs, use the exact spelling of the symbol as displayed on the GPS. If the GPS abbreviates words, spell them out.
		/// 'type' - Type (classification) of the waypoint. Accuracy Information
		/// 'fix' - Type of GPX fix.
		/// 'sat' - Number of satellites used to calculate the GPX fix.
		/// 'hdop' - Horizontal dilution of precision.
		/// 'vdop' - Vertical dilution of precision.
		/// 'pdop' - Position dilution of precision.
		/// 'ageofdgpsdata' - Number of seconds since last DGPS update.
		/// 'dgpsid' - ID of DGPS station used in differential correction. Routes Tracks Extensions (GPX 1.1 only) Version (not stored in SQL db) Creator (not stored in SQL db) 
		/// </summary>
		/// <param name="gpxPath"></param>
		public TrailSummary(string gpxPath)
		{
			this._gpxPath = gpxPath;
			GetGPXData();
		}
		/// <summary>
		/// Default constructor does nothing.
		/// </summary>
		public TrailSummary()
		{ }

		protected virtual void GetElevLegend()
		{
			StringBuilder sb = new StringBuilder();
			_totalSpan = _highPoint.ele - _lowPoint.ele;
			_sectionSpan = _totalSpan / 5;
			/*  Elevation legend format */
			sb.Append("<br><span style=\"color:#CC00CC;\">" + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 4), 0) + " - " + Convert.ToString(_highPoint)) + "</span>");
			sb.Append("<br><span style=\"color:#0034CC;\">" + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 3), 0)) + " - " + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 4), 0)) + "</span>");
			sb.Append("<br><span style=\"color:#00CC68;\">" + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 2), 0)) + " - " + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 3), 0)) + "</span>");
			sb.Append("<br><span style=\"color:#98CC00;\">" + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 1), 0)) + " - " + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 2), 0)) + "</span>");
			sb.Append("<br><span style=\"color:#CC0000;\">" + Convert.ToString(_lowPoint.ele) + " - " + Convert.ToString(Math.Round(_lowPoint.ele + (_sectionSpan * 1), 0)) + "</span>");
			this._elevLegend = sb.ToString();
		}

		/////// <summary>
		/////// This is a method to update waypoints which have been edited on page.aspx.
		/////// </summary>
		////public virtual void UpdateGPXFile()
		////{
		////    if (_gpxDataSet.Tables.Contains("wpt"))
		////    {
		////        for (int i = 0; i < _waypointArray.Count; i++)
		////        {
		////            GPXLib.wptType wp = _waypointArray[i];
		////            //Make sure each column exists before trying to write to it.  Some XML files may not have some elements.
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("ele"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["ele"] = wp.Ele;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("time"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["time"] = wp.Time;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("name"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["name"] = wp.Name;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("cmt"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["cmt"] = wp.Cmt;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("desc"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["desc"] = wp.Desc;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("sym"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["sym"] = wp.Sym;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("type"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["type"] = wp.Type;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("id"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["id"] = wp.ID;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("lat"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["lat"] = wp.Lat;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("lon"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["lon"] = wp.Lon;
		////            }
		////            if (_gpxDataSet.Tables["wpt"].Columns.Contains("gpx_Id"))
		////            {
		////                _gpxDataSet.Tables["wpt"].Rows[i]["gpx_Id"] = wp.GpxId;
		////            }
		////            foreach (DataRow myRow in _gpxDataSet.Tables["wpt"].Rows[i].GetChildRows("wpt_extensions"))
		////            {
		////                myRow.GetChildRows("extensions_label")[0]["label_text"] = wp.Label.Label_Text;
		////            }
		////        }
		////        _gpxDataSet.WriteXml(HttpContext.Current.Server.MapPath(Const.GPX_TRAIL_FOLDER + _gpxPath));
		////    }
		////    else // adding a new waypoint
		////    {
		////        XmlDocument xmlDoc;
		////        XmlNode metadataNode;
		////        XmlNode gpxNode;
		////        String xpathExpr;

		////        xmlDoc = new XmlDocument();
		////        xmlDoc.Load(HttpContext.Current.Server.MapPath(Const.GPX_TRAIL_FOLDER + _gpxPath));

		////        XmlNamespaceManager nsmgr = new XmlNamespaceManager(xmlDoc.NameTable);
		////        nsmgr.AddNamespace("ab", "http://www.topografix.com/GPX/1/1");

		////        xpathExpr = "//ab:gpx";
		////        gpxNode = xmlDoc.SelectSingleNode(xpathExpr, nsmgr);

		////        xpathExpr = "//ab:metadata";
		////        metadataNode = xmlDoc.SelectSingleNode(xpathExpr, nsmgr);

		////        List<XmlElement> elements = new List<XmlElement>();

		////        for (int i = 0; i < _waypointArray.Count; i++)
		////        {
		////            Waypoint wp = _waypointArray[i];
		////            XmlElement wpt = xmlDoc.CreateElement("", "wpt", "http://www.topografix.com/GPX/1/1");
		////            wpt.SetAttribute("lat", wp.Lat.ToString());
		////            wpt.SetAttribute("lon", wp.Lon.ToString());

		////            XmlElement ele = xmlDoc.CreateElement("", "ele", "http://www.topografix.com/GPX/1/1");
		////            XmlElement time = xmlDoc.CreateElement("", "time", "http://www.topografix.com/GPX/1/1");
		////            XmlElement name = xmlDoc.CreateElement("", "name", "http://www.topografix.com/GPX/1/1");
		////            XmlElement cmt = xmlDoc.CreateElement("", "cmt", "http://www.topografix.com/GPX/1/1");
		////            XmlElement desc = xmlDoc.CreateElement("", "desc", "http://www.topografix.com/GPX/1/1");
		////            XmlElement sym = xmlDoc.CreateElement("", "sym", "http://www.topografix.com/GPX/1/1");
		////            XmlElement type = xmlDoc.CreateElement("", "type", "http://www.topografix.com/GPX/1/1");

		////            ele.InnerText = wp.Ele.ToString();
		////            time.InnerText = wp.Time.ToString("yyyy-MM-ddThh:mm:ss");
		////            name.InnerText = wp.Name;
		////            cmt.InnerText = wp.Cmt;
		////            desc.InnerText = wp.Desc;
		////            sym.InnerText = "Flag, Red";
		////            type.InnerText = "Flag, Red";

		////            XmlElement extensions = xmlDoc.CreateElement("", "extensions", "http://www.topografix.com/GPX/1/1");
		////            XmlElement label = xmlDoc.CreateElement("", "label", "http://www.topografix.com/GPX/gpx_overlay/0/3");

		////            XmlElement label_text = xmlDoc.CreateElement("", "label_text", "http://www.topografix.com/GPX/gpx_overlay/0/3");
		////            label_text.InnerText = wp.Label.Label_Text;

		////            label.AppendChild(label_text);
		////            extensions.AppendChild(label);

		////            wpt.AppendChild(ele);
		////            wpt.AppendChild(time);
		////            wpt.AppendChild(name);
		////            wpt.AppendChild(cmt);
		////            wpt.AppendChild(desc);
		////            wpt.AppendChild(sym);
		////            wpt.AppendChild(type);
		////            wpt.AppendChild(extensions);
		////            // insert waypoints in inverse order so that newest are at top of file.
		////            elements.Insert(0, wpt);
		////        }

		////        //root.AppendChild(wpt);
		////        for (int i = 0; i < elements.Count; i++)
		////        {
		////            gpxNode.InsertAfter(elements[i], metadataNode);
		////        }

		////        xmlDoc.Save(HttpContext.Current.Server.MapPath(Const.GPX_TRAIL_FOLDER + GPXPath));
		////    }
		////}
		protected virtual System.Drawing.Color GetRGBColor(decimal elevation)
		{
			Color myColor = new Color();
			if (elevation >= _lowPoint.ele && elevation < _lowPoint.ele + (_sectionSpan * 1))
			{
				//return "#CC0000";
				myColor = Color.FromArgb(204, 0, 0);
			}
			else if (elevation >= _lowPoint.ele + (_sectionSpan * 1) && elevation < _lowPoint.ele + (_sectionSpan * 2))
			{
				//return "#98CC00";
				myColor = Color.FromArgb(152, 204, 0);
			}
			else if (elevation >= _lowPoint.ele + (_sectionSpan * 2) && elevation < _lowPoint.ele + (_sectionSpan * 3))
			{
				//return "#00CC68";
				myColor = Color.FromArgb(0, 152, 104);
			}
			else if (elevation >= _lowPoint.ele + (_sectionSpan * 3) && elevation < _lowPoint.ele + (_sectionSpan * 4))
			{
				//return "#0034CC";
				myColor = Color.FromArgb(0, 52, 204);
			}
			else // highest points
			{
				//return "#CC00CC";
				myColor = Color.FromArgb(204, 0, 204);
			}
			return myColor;
		}

		public virtual void GetGPXData()
		{
			if (System.IO.File.Exists(GPXPath))
			{

				XmlSerializer xs = new XmlSerializer(typeof(GPXLib.gpxType));
				XmlReader xr = XmlReader.Create(GPXPath);

				GPXLib.gpxType gpx = (GPXLib.gpxType)xs.Deserialize(xr);

				List<wptType> combinedTrackPoints = new List<wptType>();
				foreach (GPXLib.trkType trk in gpx.trk)
				{
					for (int i = 0; i < trk.trkseg.Length; i++)
					{
						GPXLib.trksegType trkseg = trk.trkseg[i];
						combinedTrackPoints.AddRange(trkseg.trkpt);
					}
				}

				decimal startPointLat = combinedTrackPoints.First().lat;
				decimal startPointLon = combinedTrackPoints.First().lon;
				decimal endPointLat = combinedTrackPoints.Last().lat;
				decimal endPointLon = combinedTrackPoints.Last().lon;

				//double startPointLat = Convert.ToDouble(_gpxDataSet.Tables["trkpt"].Rows[0]["lat"]);
				//double startPointLon = Convert.ToDouble(_gpxDataSet.Tables["trkpt"].Rows[0]["lon"]);
				//double endPointLat = Convert.ToDouble(_gpxDataSet.Tables["trkpt"].Rows[_gpxDataSet.Tables["trkpt"].Rows.Count - 1]["lat"]);
				//double endPointLon = Convert.ToDouble(_gpxDataSet.Tables["trkpt"].Rows[_gpxDataSet.Tables["trkpt"].Rows.Count - 1]["lon"]);

				// accumulated distance in miles from start
				double runDistFromStart = 0;
				// accumulated elevation change in meters from start
				double runElevChange = 0;

				// Get trackpoints
				for (int i = 0; i + 1 < combinedTrackPoints.Count(); i++)
				{
					// this is the current trackpoint
					GPXLib.wptType wrCurr = combinedTrackPoints[i];

					// this is the previous trackpoint, get ready to use it if needed
					GPXLib.wptType wrPrev;

					GPXLib.wptType tp = new GPXLib.wptType();
					tp.ele = Conversions.MetersToFeet(Convert.ToDecimal(wrCurr.ele));
					tp.lat = Convert.ToDecimal(wrCurr.lat);
					tp.lon = Convert.ToDecimal(wrCurr.lon);
					//DateTime tpTime;
					if (wrCurr.time != null)
					{
						tp.time = Convert.ToDateTime(wrCurr.time);
					}
					// this type does not have an RGBColor. Can create a new derived type if needed.
					////tp.RGBColor = System.Drawing.Color.Red;

					// Calculate distance from current point to beginning and end of track.  I don't think this is calculating correctly since it calculates "as crow flies", so commenting it out.
					//tp.DistanceToEnd = CDistanceBetweenLocations.Calc(endPointLat, endPointLon, Convert.ToDouble(wrCurr["lat"]), Convert.ToDouble(wrCurr["lon"]));
					if (i > 0)
					{
						// this is the previous trackpoint
						wrPrev = combinedTrackPoints[i - 1];
						// accumulate the running distance from start

						//runDistFromStart += TrackPoint.GetDistanceBetween(tp, _trackPointArray[i-1]);

						runDistFromStart += DistanceCalculator.Calc(Convert.ToDouble(wrPrev.lat), Convert.ToDouble(wrPrev.lon), Convert.ToDouble(wrCurr.lat), Convert.ToDouble(wrCurr.lon));
						// assign the current trackpoint distance from start to running total distance from start
						////tp.DistanceFromStart = runDistFromStart;

						// add to the running elevation change if ascent
						decimal rise = tp.ele - Conversions.MetersToFeet(wrPrev.ele);
						if (rise > 0)
						{
							runElevChange += (double)rise;
						}
					}

					_trackPointArray.Add(tp);
					// get high and low point while traversing the trackpoints
					if (_lowPoint.ele == -9999 ||
						_lowPoint == null ||
						_lowPoint.ele > tp.ele)
					{
						_lowPoint.ele = tp.ele;
					}
					if (_highPoint == null ||
						_highPoint.ele < tp.ele)
					{
						_highPoint.ele = tp.ele;
					}
				}

				this._totalAscent = runElevChange;
				this._totalDistance = runDistFromStart;
                if (gpx.wpt != null)
                {
                    this._waypointArray = gpx.wpt.ToList();
                }
			}
		}

		/// <summary>
		/// This method is used to find trackpoints closest to where pictures were taken. Returns nearest trackpoint in time AFTER the passed in time.
		/// </summary>
		/// <param name="myTime">The time to compare against trackpoint times.</param>
		/// <returns></returns>
		public GPXLib.wptType GetClosestTrackpoint(DateTime myTime)
		{
			long tmpPointDif = 999999999;
			long tmpPointDifCur = 0;
			//TrackPoint tmpTrackPoint = new TrackPoint();
			for (int j = 0; j < _trackPointArray.Count; j++)
			{
				GPXLib.wptType tp = (GPXLib.wptType)_trackPointArray[j];
				// run through the trackpoints until we find the one we're looking for
				if (tp.time.Ticks < myTime.Ticks)
				{
					tmpPointDifCur = Math.Abs(tp.time.Ticks - myTime.Ticks);
					if (tmpPointDifCur < tmpPointDif)
					{
						tmpPointDif = tmpPointDifCur;
						//tmpTrackPoint = tp;
					}
				}
				else if (j > 0)
				{
					return _trackPointArray[j - 1];
				}
			}
			return new GPXLib.wptType();
		}


		#region Public Accessor Methods

		public virtual string GPXPath
		{
			get { return _gpxPath; }
			set { _gpxPath = value; }
		}
		public GPXLib.wptType HighPoint
		{
			get { return _highPoint; }
			set { _highPoint = value; }
		}

		public GPXLib.wptType LowPoint
		{
			get { return _lowPoint; }
			set { _lowPoint = value; }
		}
		/// <summary>
		///  Elevation difference in meters from lowest to highest point (not accum. elev. change). Used for color coding the track on a map.
		/// </summary>
		public decimal TotalSpan
		{
			get { return _totalSpan; }
			set { _totalSpan = value; }
		}
		/// <summary>
		///  Elevation difference in meters for each section of the legend used when color coding a track on a map.
		/// </summary>
		public decimal SectionSpan
		{
			get { return _sectionSpan; }
			set { _sectionSpan = value; }
		}
		/// <summary>
		///  Total distance in miles from start to end of trail, based on total of distances between each trackpoint.
		/// </summary>
		public double TotalDistance
		{
			get { return _totalDistance; }
			set { _totalDistance = value; }
		}
		/// <summary>
		/// Total elevation change based on diff between each trackpoint
		/// </summary>
		public double TotalElevChange
		{
			get { return _totalElevChange; }
			set { _totalElevChange = value; }
		}
		public double TotalAscent
		{
			get { return _totalAscent; }
			set { _totalAscent = value; }
		}
		public virtual string ElevLegend
		{
			get { return _elevLegend; }
			set { _elevLegend = value; }
		}
		public virtual GPXLib.gpxType GPXDataSet
		{
			get { return _gpxDataSet; }
			set { _gpxDataSet = value; }
		}
		public virtual List<GPXLib.wptType> TrackPointArray
		{
			get { return _trackPointArray; }
			set { _trackPointArray = value; }
		}
		public virtual List<GPXLib.wptType> WaypointArray
		{
			get { return _waypointArray; }
			set { _waypointArray = value; }
		}
		#endregion
	}
}
